unit richer;

interface

uses ComCtrls,Controls,Windows,Messages;

CONST
{Constants to allow right, centre and decimal tabs}
{also reputed to help with rendering Arabic and Thai fonts..
 if anyone is interested in that!}
EM_SETTYPOGRAPHYOPTIONS	= (WM_USER + 202);
EM_GETTYPOGRAPHYOPTIONS	= (WM_USER + 203);
TO_ADVANCEDTYPOGRAPHY =	1;
TO_SIMPLELINEBREAK    = 2;

{Messages to increase/decrease apparent size of text}
EM_GETZOOM 		= (WM_USER + 224);
EM_SETZOOM 		= (WM_USER + 225);



TYPE
  TRicherEdit=Class(TRichEdit)
  protected
   {the critical stuff is in here}
   procedure CreateParams(var Params: TCreateParams); override;
  private
   Flibhandle :integer;
  public
   Destructor Destroy; Override;
   {re-implementing FindText, to remove default behaviour of
    searching backwards, thanks M$!}
   function FindText(const SearchStr: string;
        StartPos, Length: Integer; Options: TSearchTypes): Integer;
   {Not in demo, but a change to the Print routine is also needed
   if you plan to call it, otherwise it will continue throwing pages
   until you throw the printer out of the window..}
   procedure Print(const Caption:string);
  end;

implementation

uses Richedit,
     Printers {needed in print method}
     ;

procedure TRicherEdit.CreateParams(var Params: TCreateParams);
const
  RichEditModuleName = 'RICHED20.DLL';
begin
  FLibHandle := LoadLibrary(RichEditModuleName);
  if (FLibHandle > 0) and (FLibHandle < HINSTANCE_ERROR) then FLibHandle := 0;

  inherited CreateParams(Params);

  CreateSubClass(Params, 'RICHEDIT20A');
end;

Destructor TRicherEdit.Destroy;
begin
    if Flibhandle <> 0 then FreeLibrary(Flibhandle);
    inherited;
end;


function TRicherEdit.FindText(const SearchStr: string;
  StartPos, Length: Integer; Options: TSearchTypes): Integer;
var
  Find: TFindText;
  Flags: Integer;
begin
  Find.chrg.cpMin := StartPos;
  Find.chrg.cpMax := Find.chrg.cpMin + Length;

  Flags:=1;    {Search Forwards: Richedit v2 on, ignored by v1}
  if stWholeWord in Options then Flags := Flags or FT_WHOLEWORD;
  if stMatchCase in Options then Flags := Flags or FT_MATCHCASE;
  Find.lpstrText := PChar(SearchStr);
  Result := SendMessage(Handle, EM_FINDTEXT, Flags, LongInt(@Find));
end;

procedure TRicherEdit.Print(const Caption: string);
      function GetTextLenNoLinebreaks:integer;
      {This is the essential change to fix the endless page-throwing,
       so I've put it here out of line; below, call this function
       instead of GetTextLength;
       The problem is that internally the class 1 Richedit uses CR/LF pairs,
       but the Class 2 etc uses a single character;
       so for example, SelStart, when the caret is at the end of the text,
       is considerably less than length(text)}
      var gtlx:integer;
          len:integer;
      begin
         {I've seen some implementations (3rd party Richedits) that use
          GTL_PRECISE OR CP_AP, here
          but when I tried this it always returned zero ;
          The following works for me}
         gtlx:=GTL_NUMCHARS or CP_ACP;
         len:=Perform(EM_GETTEXTLENGTHEX,integer(@gtlx),0);
         result:=len;
      end;
var
  Range: TFormatRange;
  LastChar, MaxLen, LogX, LogY, OldMap: Integer;
begin
  FillChar(Range, SizeOf(TFormatRange), 0);
  with Printer, Range do {! from the VCL; of course you would NEVER write like this}
  begin
    Title := Caption;
    BeginDoc;
    hdc := Handle;
    hdcTarget := hdc;
    LogX := GetDeviceCaps(Handle, LOGPIXELSX);
    LogY := GetDeviceCaps(Handle, LOGPIXELSY);
    if IsRectEmpty(PageRect) then
    begin
       {while here I also modified the page settings to give a
        half inch margin all round : if not wanted,
        remove all the (log? div 2) elements}
      rc.right := (PageWidth-(LogX div 2)) * 1440 div LogX;
      rc.bottom := (PageHeight-(LogY div 2)) * 1440 div LogY;
      rc.left := (LogX div 2) * 1440 div LogX;
      rc.top := (LogY div 2) * 1440 div LogY;
    end
    else begin
      rc.left := PageRect.Left * 1440 div LogX;
      rc.top := PageRect.Top * 1440 div LogY;
      rc.right := PageRect.Right * 1440 div LogX;
      rc.bottom := PageRect.Bottom * 1440 div LogY;
    end;
    rcPage := rc;
    LastChar := 0;
    {The important fix is here}
    MaxLen := GetTextLenNoLinebreaks;
    {** essential not to use gettextlen here, which loops forever **}

    {The rest of the routine is just as in VCL}
    chrg.cpMax := -1;
    // ensure printer DC is in text map mode
    OldMap := SetMapMode(hdc, MM_TEXT);
    SendMessage(Handle, EM_FORMATRANGE, 0, 0);    // flush buffer
    try
      repeat
        chrg.cpMin := LastChar;
        LastChar := SendMessage(Self.Handle, EM_FORMATRANGE, 1, Longint(@Range));
        if (LastChar < MaxLen) and (LastChar <> -1) then NewPage;
      until (LastChar >= MaxLen) or (LastChar = -1);
      EndDoc;
    finally
      SendMessage(Handle, EM_FORMATRANGE, 0, 0);  // flush buffer
      SetMapMode(hdc, OldMap);       // restore previous map mode
    end;
  end;
end;



END.
